#!/usr/bin/env lua

os = require('os')
lfs = require('lfs')
dir = require('pl.dir')
stringx = require('pl.stringx')
utils = require('pl.utils')
pretty = require('pl.pretty')
path = require('pl.path')
file = require('pl.file')
app = require('pl.app')
tablex = require('pl.tablex')
tz = require('luatz')
require('pl.strict')

-- Global variable declarations
users = {}		-- Table of user tables, maintained in a sorted order
users_by_name = {}	-- Mapping from usernames to user tables, kept in synch with users
current_user = nil	-- Pointer to the current user, used by sub-commands of "user" command
online_only = false	-- Boolean option, whether or not to display only logged-in users.
colours = true		-- Boolean, controls whether to use ANSI colours

-- Utility functions

-- This is used for date parsing and hopefully can be replaced
-- at some point by an actual date parsing library.

days = {
	Sun=true,
	Mon=true,
	Tue=true,
	Wed=true,
	Thu=true,
	Fri=true,
	Sat=true
}

month_to_int = {
	Jan=1,
	Feb=2,
	Mar=3,
	Apr=4,
	May=5,
	Jun=6,
	Jul=7,
	Aug=8,
	Sep=9,
	Oct=10,
	Nov=11,
	Dec=12
}

_COLOURS = {
	red=31,
	green=32,
	yellow=33,
	blue=34,
	magenta=35,
	cyan=36,
}

function get_editor()
	-- Find this user's preferred editor, or default to nano.
	-- Could be improved, e.g. by checking that nano actually exists.
	local editor = os.getenv("VISUAL")
	if not editor then editor = os.getenv("EDITOR") end
	if not editor then editor = "nano" end
	return editor
end

function colourise(colour, str)
	if not colours or _COLOURS[colour] == nil then return str end
	return string.char(27) .. "[" .. tostring(_COLOURS[colour]) .. "m" .. str .. string.char(27) .. "[0m"
end

function getchar()
	-- Read a single character from stdin, without requiring the
	-- user to hit 'enter' to return.
	os.execute("/bin/stty -icanon")
	local char = io.read(1)
	os.execute("/bin/stty icanon")
	return char
end

function dispatch_loop(dispatch_table, quit_char, prompt_func, err_str)
	-- An infinite loop of displaying a prompt, reading a single char,
	-- using that char to find a function and calling it, which runs
	-- until the user enters the one magic quit char.
	repeat
		-- Show prompt and read 1 char from keyboard
		io.write(prompt_func())
		local c = getchar()
		if c ~= "\n" then io.write("\n") end
		-- Use char as index into dispatch table
		if c == "\n" then
			-- pass
		elseif dispatch_table[c] == nil then
			print(err_str)
		else
			dispatch_table[c]()
		end
	until c == quit_char
end

function strptime(time_tab)
	-- Hideous fragile likely broken date parsing
	-- time_tab is an array of strings
	-- time_tab[1] = abbreviated day of week
	-- time_tab[2] = abbreviated month
	-- time_tab[3] = day of month
	-- time_tab[4] = some kind of time
	local last = {}
	last.month = month_to_int[time_tab[2]]
	last.day = tonumber(time_tab[3])
	-- Fail early if something has gone wrong
	if not last.month or not last.day then return nil end
	local now = os.date("*t")
	last.year = now.year
	if last.month > now.month then
		last.year = now.year - 1
	else
		last.year = now.year
	end
	return last
end

-- Internals

function update_local_users()

	-- Iterate over the main user table, skipping over remote users.
	-- Parse the output of Unix commands `users` and `last` to update
	-- local users' tables to reflect their login status and times.

	local logged_in_users = {}
	local err, ret, stdout, stderr = utils.executeex("users")
	for _,user in pairs(stringx.split(stdout)) do
		logged_in_users[user] = true
	end
	for _,user in pairs(users) do
		if user.is_remote then goto continue end
		user.logged_in = logged_in_users[user.username]
		-- An odd convention which facilitates the user sorting
		-- we want: currently logged in users are represented as
		-- having "last logged in" just before midnight today.
		if user.logged_in then
			local last = os.date("*t")
			last.hour = 23
			last.min = 59
			last.sec = 59
			user.last_login = os.time(last)
			goto continue
		end
		local err, ret, stdout, stderr = utils.executeex("last -n 1 " .. user.username .. " | head -n 1", false)
		if stdout == "\n" then
			user.last_login = 0
		else
			local bits = stringx.split(stdout)
			local i = 1
			while i <= #bits do
				if days[bits[i]] then break end
				i = i+1
			end
			if i == #bits then
				-- Couldn't find the start of a datetime
				return 0
			end
			local time_bits = tablex.sub(bits, i, i+4)
			local last = strptime(time_bits)
			if last then
				user.last_login = os.time(last)
			else
				user.last_login = 0
			end
		end
		:: continue ::
	end
end

recognised_headers = {}
recognised_headers["Project:"] = true
recognised_headers["Full name:"] = true
recognised_headers["Timezone:"] = true

function populate_remote_user_table(user)

	-- Treat the passed in table as a user table.
	-- At a bare minimum it should have username set to a remote
	-- username, e.g user@host.
	-- Try to remotely finger that user, and extract as much data
	-- as we can from the result, e.g. login status and times,
	-- plan, timezone, etc.  Store it in the user table.
	-- This can be used to initialise a new remote user table OR
	-- to update an existing one by overwriting changed values.

	if user.username == nil then return nil end
	-- Try to do the remote finger
	local err, ret, stdout, stderr = utils.executeex("finger " .. user.username)
	if ret ~= 0 or stringx.count(stdout, "no such user") ~=0 then
		return nil
	end
	-- This should already be true, but let's make sure.
	user.is_remote = true
	-- Try to parse stuff out
	-- First go through word by word and extract all we can
	local fields = {}
	local field = nil
	local bits = {}
	stdout = string.gsub(stdout, "Last login", "Last_login:")
	stdout = string.gsub(stdout, "Last_login::", "Last_login:") -- Hacky!
	stdout = string.gsub(stdout, "On since", "On_since:")
	stdout = string.gsub(stdout, "Full name:", "Full_name:")
	local words = stringx.split(stdout)
	for _,word in ipairs(words) do
		if stringx.endswith(word, ":") then
			-- New field
			if field then
				fields[field] = stringx.join(" ", bits)
			end
			field = string.sub(word, 1, string.len(word)-1)
			bits = {}
		else
			table.insert(bits, word)
		end
	end
	if field then
		fields[field] = stringx.join(" ", bits)
	end
	if fields.Plan ~= nil and fields.Plan ~= "" then user.remote_plan = fields.Plan end
	if fields.Timezone then user.tz = fields.Timezone end
	if fields.On_since then
		-- User is logged in
		user.logged_in = true
		local last = os.date("*t")
		last.hour = 23
		last.min = 59
		last.sec = 59
		user.last_login = os.time(last)
	end
	if fields.Last_login then
		-- User is logged out, but we can see when they
		-- were last logged in
		user.logged_in = false
		local bits = stringx.split(fields.Last_login)
		local time_bits = tablex.sub(bits, 1, 3)
		if #time_bits == 0 then
			-- Parse error
			user.last_login = 0
		else
			local last = strptime(time_bits)
			user.last_login = os.time(last)
		end
	end
	return stdout
end

function check_perms()
	-- Check the permissions on the executing user's .plan, .project and .tz
	-- files, and offer to fix them if they aren't world readable.
	local files = { ".project", ".plan", ".tz" }
	for _,file in pairs(files) do
		local full_path = path.join(os.getenv("HOME"), file)
		if path.exists(full_path) then
			local perms = lfs.attributes(full_path).permissions
			if string.sub(perms,4,4) ~= "r" or string.sub(perms,7,7) ~= "r" then
				print("Hey, your " .. file .. " file isn't world readable!")
				io.write("Would you like me to fix it for you? (y/n) ")
				local resp = string.lower(getchar())
				print("")
				if resp == "y" then
					os.execute("chmod og+r " .. full_path)
				end
			end
		end
	end
end

function add_local_users()
	-- Parse /etc/passwd and build user tables for all local users who survive
	-- a filtering function which tries to exclude daemons, etc.  Also ignores
	-- users whose names are specified in the executing users's personal ignore
	-- list.
	local to_ignore = load_ignored_users()
	io.input("/etc/passwd")
	for line in io.lines() do
		if stringx.startswith(line, "#") then goto continue end
		local bits = stringx.split(line,":")
		local user = {}
		user.is_remote = false
		user.username = bits[1]
		if to_ignore[user.username] then goto continue end
		user.uid = tonumber(bits[3])
		user.gid = tonumber(bits[4])
		user.gecos = bits[5]
		user.home = bits[6]
		user.shell = bits[7]
		user.tz = nil
		local tz_file = path.join(user.home, ".tz")
		if path.exists(tz_file) then user.tz = stringx.strip(file.read(tz_file)) end

		if user_filter(user) then
			table.insert(users, user)
		end
		:: continue ::
	end
	io.input():close()
	io.input(io.stdin)
end

function load_ignored_users()
	-- Return a set of usernames specified in the executing user's file of
	-- local users to ignore.  If that file doesn't exist, return an empty
	-- set.
	local ignored = {}
	local ignored_user_file = path.join(os.getenv("HOME"), ".fellowsh", "ignored_locals")
	if not path.exists(ignored_user_file) then return ignored end
	for user in io.lines(ignored_user_file) do
		ignored[user] = true
	end
	return ignored
end

function add_remote_users()
	-- Read the executing user's file of remote users to follow.  For each
	-- one, insert a skeletal user table into the users table.  These must then
	-- be fleshed out by a call to update_remote_users().  Note also that this
	-- function does not update users_by_name.  That is handled in the main
	-- program initialisation, which is the only place this is called.
	local remote_user_file = path.join(os.getenv("HOME"), ".fellowsh", "remote_users")
	if not path.exists(remote_user_file) then return end
	for remote in io.lines(remote_user_file) do
		local user = {}
		user.username = remote
		user.is_remote = true
		user.logged_in = false
		user.last_login = 0
		user.remote_plan = "No remote plan."
		table.insert(users, user)
	end
end

function get_plan_line(user)
	-- Return a possibly truncated plan line to display in the main user listing.
	if user.is_remote then
		if string.len(user.remote_plan) <= 48 then
			return user.remote_plan
		else
			return string.sub(user.remote_plan, 1, 48-3) .. "..."
		end
	end
	if path.exists(path.join(user.home, ".nofinger")) then return "No plan." end
	if not path.exists(path.join(user.home, ".plan")) then return "No plan." end
	local fp, err = io.open(path.join(user.home, ".plan"), "r")
	if fp ~= nil then
		local line = fp:read()
		fp:close()
		if string.len(line) > 48 then
			line = string.sub(line, 1, 48-3) .. "..."
		end
		return line
	else
		return "~/.plan exists but is not world-readable."
	end
end

function update_remote_users()
	-- Do a remote finger on each remote user and update their info.
	local first = true
	for _,user in pairs(users) do
		if not user.is_remote then goto continue end
		if first then
			first = false
			io.write("Polling remote hosts")
		else
			io.write(".")
		end
		io.flush()
		populate_remote_user_table(user)
		:: continue ::
	end
	-- Print a new line if we've been printing dots
	if not first then print("") end
end

function user_filter(user)
	-- Return true if a user appears to be a "real human", and false if
	-- it is a system account or something like that.
	local max_uid = 500 -- Starting uid for humans on Linux
	-- Check uname and adjust max_uid for other platforms
        local success, ret, stdout, stderr = utils.executeex("uname -s", false)
        if success and string.find(stdout,"FreeBSD") then max_uid = 1000 end
	if user.uid < max_uid then return false end
	-- Look to shell to spot users who aren't real users
	if user.shell == "/bin/false" then return false end
	if user.shell == "/usr/sbin/nologin" then return false end
	return true
end

function user_sort_function(x,y)
	-- Table sorting function for the main user listing.
	if x.last_login == y.last_login then
		return x.username < y.username
	else
		return x.last_login > y.last_login
	end
end

-- Commands

function do_edit_gecos()
	-- Just launch `chfn` to let the user update their GECOS
	print("Editing your GECOS information via `chfn`...")
	os.execute("chfn")
end

function do_help()
	-- Display a help message
	print([[Fellowsh commands:

[a]dd...................Add remote user to roster
[c]olour................Toggle coloured output
[e]dit..................Edit your GECOS info
[g]oto..................Goto user, same as [u]ser
[h]elp..................Display this help message
[i]gnore................Ignore local user
[l]ist..................List all users
[o]nline................Toggle online-only view
[p]lan..................Edit your .plan file
[P]roject...............Edit your .project file
[q]uit..................Quit
[r]efresh...............Refresh list
[t]imezone..............Update your timezone file
[u]ser..................Display user info]])
end

function do_add()
	-- Attempt to add a remote user
	io.write("Remote user: ")
	local remote = io.read()
	if remote == "" then
		print("Cancelling add.")
		return
	end
	-- Don't allow duplications
	if users_by_name[remote] then
		print("User already in roster!")
		return
	end
	-- Make sure this is really a remote user
	if stringx.count(remote, "@") == 0 then
		print("Not a remote username!")
		return
	end
	-- Do the remote finger and add the user to the users table
	local user = {}
	user.username = remote
	user.is_remote = true
	user.logged_in = false
	user.last_login = 0
	user.remote_plan = "No remote plan."
	local finger_output = populate_remote_user_table(user)
	if finger_output then
		table.insert(users, user)
		users_by_name[user.username] = user
		table.sort(users, user_sort_function)
	else
		print("Adding remote user failed.")
		return
	end
	-- Display the output of `finger`
	io.write(finger_output)
	-- Save this remote user's name@host to the persistent list
	os.execute("mkdir -p ~/.fellowsh/")
	local fp = io.open(path.join(os.getenv("HOME"), ".fellowsh", "remote_users"),"a")
	fp:write(remote .. "\n")
	fp:close()
end

function do_colour()
	-- Toggle coloured output
	colours = not colours
end

function do_ignore()
	-- Add a local user to the persistent list of local users to ignore.
	io.write("User to ignore: ")
	local username = io.read()
	if username == "" then
		print("Cancelling ignore.")
		return
	end
	if not users_by_name[username] then
		print("No such user!")
		return
	end
	if users_by_name[username].is_remote then
		print(username .. " is a remote user!")
		return
	end

	local index_to_rm = nil
	for i,user in pairs(users) do
		if user.username == username then
			index_to_rm = i
			break
		end
	end
	assert(users[index_to_rm].username == username)
	table.remove(users, index_to_rm)
	print("Ignoring local user " .. username .. ".")

	os.execute("mkdir -p ~/.fellowsh/")
	local fp = io.open(path.join(os.getenv("HOME"), ".fellowsh", "ignored_locals"),"a")
	fp:write(username .. "\n")
	fp:close()
end

function do_list()
	-- Display a summary list of all users, indicatin their login status, timezone
	-- and plan.  This is the main "view" of the the program.

	-- Handle an empty list (e.g. remote only mode with no remote users added yet)
	if #users == 0 then
		print("Ain't nobody here but you!")
		return
	end

	-- Header line
	io.write(stringx.ljust("ID",4))
	io.write(stringx.ljust("USERNAME",24))
	io.write(stringx.ljust("LAST LOGIN",18))
	io.write(stringx.ljust("LOCAL TIME",12))
	print("PLAN")
	-- Iterate over users
	for i,user in ipairs(users) do
		if online_only and not user.logged_in then goto continue end
		-- ID
		io.write(stringx.ljust(tostring(i), 4))
		-- Username
		if user.logged_in then
			io.write(colourise("green", stringx.ljust(user.username, 24)))
		else
			io.write(stringx.ljust(user.username, 24))
		end
		-- Last login
		if user.logged_in then
			io.write(stringx.ljust("Logged in now",18))
		elseif user.last_login ~= 0 then
			io.write(stringx.ljust(os.date("%a %b %d", user.last_login),18))
		else
			if user.is_remote then
				io.write(stringx.ljust("Unknown (remote)",18))
			else
				io.write(stringx.ljust("None lately",18))
			end
		end
		-- Local time
		if user.tz then
			local ok, time = pcall(tz.time_in, user.tz)
			if ok then
				time = math.floor(time)
				io.write(stringx.ljust(os.date("%X", time),12))
			else
				io.write(stringx.ljust("Invalid TZ",12))
			end
		else
			io.write(stringx.ljust("No TZ set.",12))
		end
		-- Plan
		local no_newlines = string.gsub(get_plan_line(user),"\n"," ")
		print(no_newlines)
		:: continue ::
	end
end

function do_online()
	-- Toggle online only mode
	online_only = not online_only
	do_list()
end

function do_plan()
	-- Edit user's plan file
	local plan_file = path.join(os.getenv("HOME"), ".plan")
	local editor = get_editor()
	os.execute(editor .. " " .. plan_file)
end

function do_project()
	-- Edit user's plan file
	local project_file = path.join(os.getenv("HOME"), ".project")
	local editor = get_editor()
	os.execute(editor .. " " .. project_file)
end

function do_refresh()
	-- Poll all remote hosts to update user information.
	update_local_users()
	update_remote_users()
	table.sort(users, user_sort_function)
	do_list()
end

function do_timezone()
	local db_path = "/usr/share/zoneinfo"
	if not path.exists("/usr/share/zoneinfo/") then
		print("Sorry, can't find local timezone database...")
		return
	end
	local country = nil
	local city = nil
	while true do
		local countries = dir.getdirectories(db_path)
		table.sort(countries)
		for i,country in ipairs(countries) do
			io.write(stringx.ljust("["..tostring(i).."]", 5) ..
				stringx.ljust(path.basename(country),18))
			if i % 4 == 0 then print("") end
		end
		print("["..tostring(#countries+1).."] Cancel")
		io.write("Choose your country/region: ")
		local index = tonumber(io.read())
		country = countries[index]
		if country then break end
		if index == #countries+1 then return end
		print("Invalid choice.")
	end
	while true do
		local cities = dir.getfiles(country)
		table.sort(cities)
		for i,city in ipairs(cities) do
			io.write(stringx.ljust("["..tostring(i).."]", 5) ..
				stringx.ljust(path.basename(city),18))
			if i % 4 == 0 then print("") end
		end
		print("["..tostring(#cities+1).."] Cancel")
		io.write("Choose your city: ")
		local index = tonumber(io.read())
		city = cities[index]
		if city then break end
		if index == #cities+1 then return end
		print("Invalid choice.")
	end
	local tz_string = path.join(path.basename(country), path.basename(city))
	io.write("Is this your local time? ")
	local time = tz.time_in(tz_string)
	io.write(os.date("%c", time).." [y/n] ")
	local resp = nil
	while true do
		resp = string.lower(getchar())
		print("")
		if resp == "y" or resp == "n" then break end
		io.write("Please hit y or n...")
	end
	if resp == "y" then
		print("Writing " .. tz_string .. " to ~/.tz file...")
		local tz_file = path.join(os.getenv("HOME"), ".tz")
		local fp = io.open(tz_file, "w")
		fp:write(tz_string.."\n")
		fp:close()
		os.execute("chmod og+r " .. tz_file)
	else
		print("Timezone set cancelled.")
	end
end
-- Sub-commands of "user" command

function do_user_finger()
	-- Finger a user
	local user = current_user
	os.execute("finger " .. current_user.username)
end

function do_user_proc()
	-- Display a user's list of running processes
	local user = current_user
	if current_user.is_remote then print("Process list not available for remote users.") return end
	print("[PROCESSES]\n")
	local err, ret, stdout, stderr = utils.executeex("ps -fu " .. user.username)
	for line in stringx.lines(stdout) do
		local bits = stringx.split(line)
		if bits[8] == "CMD" then goto continue end
		for i=8,#bits do
			io.write(bits[i] .. " ")
		end
		print()
		:: continue ::
	end
end

function do_user_talk()
	-- Launch a `talk` session with a user.
	local user = current_user
	if current_user.is_remote then print("Talk not available for remote users.") return end
	os.execute("talk " .. current_user.username)
end

function do_user_write()
	-- `write` to a user.
	local user = current_user
	if current_user.is_remote then print("Write not available for remote users.") return end
	print("Now `write`ing to " .. user.username .. ".  Press Ctrl-D when finished.")
	os.execute("write " .. current_user.username)
end

-- Actual "user" command

user_dispatch_table = {
	f=do_user_finger,
	p=do_user_proc,
	t=do_user_talk,
	w=do_user_write,
}

function do_user()
	-- "Go to" a particular user.  This will display the results of fingering
	-- them and then display a sub-prompt giving options to interact with them
	-- or view other details.
	io.write("Which user? (username or ID) ")
	local key = io.read()
	-- Try numeric index first
	local id = tonumber(key)
	current_user = nil
	if id ~= nil then
		current_user = users[id]
	else
	-- Now try username
		current_user = users_by_name[key]
	end
	if current_user == nil then
		print("No such user, use numeric ID or full username.")
		return
	end
	os.execute("finger " .. current_user.username)
	dispatch_loop(user_dispatch_table, "d", function()
		return colourise("cyan", current_user.username) ..
			" [" .. colourise("red", "f") .. "]inger " ..
			"[" .. colourise("red", "p") .. "]rocs " ..
			"[" .. colourise("red", "t") .. "]alk " ..
			"[" .. colourise("red", "w") .. "]rite " ..
			"[" .. colourise("red", "d") .. "]one > "
	end, "Unrecognised command!")
	do_list()
end

-- Main program

local help_string = [[Fellowsh usage:

  -h            Help.
                Display this message and then quit.

  -l            Local users only.
                Do not read remote users from ~/.fellowsh/remote, which makes
                fellowsh very fast as no remote servers have to be polled.

  -n            Non-interactive mode.
                Print a user listing and then immediately quit without reading
                any input from the user.

  -o            Online users only.
                Show only users who are logged in right now.  In combination
                with non-interactive mode, this turns fellowsh into a possible
                replacement for the traditional `who` command.

  -r            Remote users only.
		Do not display local users, only remote users listed in
                ~/.fellowsh/remote.]]

-- Parse command line arguments
local flags, _ = app.parse_args()
if flags.o then online_only = true end
if flags.l and flags.r then
	print("Can't set -l and -r together!")
	os.exit()
end

-- Display help and quit, if that's all we were asked to do
if flags.h then
	print(help_string)
	os.exit()
end

-- Be friendly if we're running in interactive mode
if not flags.n then
	print("Welcome to fellowsh!")
	check_perms()
end

-- Initialise user tables
if not flags.r then
	add_local_users()
end
if not flags.l then
	add_remote_users()
end
for _,user in pairs(users) do
	users_by_name[user.username] = user
end

-- The do_refresh() command will update the metadata
-- of all users, sort the user table and display it
do_refresh()

-- If we're in non-interactive mode, that's it
if flags.n then os.exit() end

-- Command dispatch loop for interactive mode
main_dispatch_table = {
	a=do_add,
	c=do_colour,
	e=do_edit_gecos,
	g=do_user,
	h=do_help,
	i=do_ignore,
	l=do_list,
	o=do_online,
	p=do_plan,
	P=do_project,
	r=do_refresh,
	t=do_timezone,
	u=do_user,
	q=function() print(
[["What we wanted to preserve was not just a good environment in which
to do programming, but a system around which fellowship could form."
                                                         -Dennis Ritchie
]]) end,
}
dispatch_loop(main_dispatch_table, "q", function() return "> " end, "Unrecognised command!")
